{% extends "!layout.html" %}



{% block extrahead %}

    <!-- Global site tag (gtag.js) - Google Analytics -->
    <script async src="https://www.googletagmanager.com/gtag/js?id=UA-166989761-1"></script>
    <script>
    window.dataLayer = window.dataLayer || [];
    function gtag(){dataLayer.push(arguments);}
    gtag('js', new Date());

    gtag('config', 'UA-166989761-1');
    </script>

    <link rel="stylesheet" href="{{ pathto('_static/style.css', 1) }} " type="text/css" />
    <link rel="stylesheet" href="{{ pathto('_static/font-awesome.css', 1) }}" type="text/css" />
    <link rel="stylesheet" href="{{ pathto('_static/font-source-code-pro.css', 1) }}" type="text/css" />
    <link rel="stylesheet" href="{{ pathto('_static/font-source-sans-pro.css', 1) }}" type="text/css" />
    <link rel="stylesheet" href="{{ pathto('_static/flag-icon.css', 1) }}" type="text/css" />
    
{{ super() }}
{% endblock %}

{# put the sidebar before the body #}
{% block sidebar1 %}{{ sidebar() }}{% endblock %}
{% block sidebar2 %}{% endblock %}
